from __future__ import annotations

import warnings
from pathlib import Path
from types import MappingProxyType
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    ClassVar,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import numpy as np
from pydantic import BaseModel

from useq._pydantic_compat import PYDANTIC2, model_dump, model_fields

if TYPE_CHECKING:
    ReprArgs = Sequence[Tuple[Optional[str], Any]]
    IncEx = set[int] | set[str] | dict[int, Any] | dict[str, Any] | None

__all__ = ["UseqModel", "FrozenModel"]

_T = TypeVar("_T", bound="FrozenModel")
_Y = TypeVar("_Y", bound="UseqModel")


class FrozenModel(BaseModel):
    if PYDANTIC2:
        model_config = {
            "populate_by_name": True,
            "extra": "ignore",
            "frozen": True,
        }

    else:

        class Config:
            allow_population_by_field_name = True
            extra = "ignore"
            frozen = True
            json_encoders: ClassVar[dict] = {MappingProxyType: dict}

    def replace(self: _T, **kwargs: Any) -> _T:
        """Return a new instance replacing specified kwargs with new values.

        This model is immutable, so this method is useful for creating a new
        sequence with only a few fields changed.  The uid of the new sequence will
        be different from the original.

        The difference between this and `self.copy(update={...})` is that this method
        will perform validation and casting on the new values, whereas `copy` assumes
        that all objects are valid and will not perform any validation or casting.
        """
        state = model_dump(self, exclude={"uid"})
        return type(self)(**{**state, **kwargs})

    if PYDANTIC2:
        # retain pydantic1's json method
        def json(
            self,
            *,
            indent: int | None = None,  # type: ignore
            include: IncEx = None,
            exclude: IncEx = None,  # type: ignore
            by_alias: bool = False,
            exclude_unset: bool = False,
            exclude_defaults: bool = False,
            exclude_none: bool = False,  # type: ignore
            round_trip: bool = False,
            warnings: bool = True,
        ) -> str:
            return super().model_dump_json(
                indent=indent,
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
                round_trip=round_trip,
                warnings=warnings,
            )

        # we let this one be deprecated
        # def dict()

    elif not TYPE_CHECKING:
        # Backport pydantic2 methods so that useq-0.1.0 can be used with pydantic1

        def model_dump_json(self, **kwargs: Any) -> str:
            """Backport of pydantic2's model_dump_json method."""
            return self.json(**kwargs)

        def model_dump(self, **kwargs: Any) -> dict[str, Any]:
            """Backport of pydantic2's model_dump_json method."""
            return self.dict(**kwargs)


class UseqModel(FrozenModel):
    def __repr_args__(self) -> ReprArgs:
        """Only show fields that are not None or equal to their default value."""
        return [
            (k, val)
            for k, val in super().__repr_args__()
            if k in model_fields(self)
            and val
            != (
                factory()
                if (factory := model_fields(self)[k].default_factory) is not None
                else model_fields(self)[k].default
            )
        ]

    @classmethod
    def from_file(cls: Type[_Y], path: Union[str, Path]) -> _Y:
        """Return an instance of this class from a file.  Supports JSON and YAML."""
        path = Path(path)
        if path.suffix in {".yaml", ".yml"}:
            import yaml

            obj = yaml.safe_load(path.read_bytes())
        elif path.suffix == ".json":
            import json

            obj = json.loads(path.read_bytes())
        else:  # pragma: no cover
            raise ValueError(f"Unknown file type: {path.suffix}")

        return cls.model_validate(obj) if PYDANTIC2 else cls.parse_obj(obj)

    @classmethod
    def parse_file(cls: Type[_Y], path: Union[str, Path], **kwargs: Any) -> _Y:
        warnings.warn(  # pragma: no cover
            "parse_file is deprecated. Use from_file instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return cls.from_file(path)  # pragma: no cover

    def yaml(
        self,
        *,
        include: Optional[Union[set, dict]] = None,
        exclude: Optional[Union[set, dict]] = None,
        by_alias: bool = False,
        exclude_unset: bool = True,  # pydantic has False by default
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        stream: Optional[IO[str]] = None,
    ) -> Optional[str]:
        """Generate a YAML representation of the model.

        Returns
        -------
        yaml : str or None
            YAML output ... If `stream` is provided, returns `None`.
        """
        from datetime import timedelta
        from enum import Enum

        import yaml

        yaml.SafeDumper.add_multi_representer(
            timedelta, lambda dumper, data: dumper.represent_str(str(data))
        )
        yaml.SafeDumper.add_multi_representer(
            Enum, lambda dumper, data: dumper.represent_str(str(data.value))
        )
        yaml.SafeDumper.add_multi_representer(
            MappingProxyType, lambda dumper, data: dumper.represent_dict(data)
        )
        yaml.SafeDumper.add_multi_representer(
            np.floating, lambda dumper, d: dumper.represent_float(float(d))
        )

        data = model_dump(
            self,
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )
        return yaml.safe_dump(data, stream=stream)



import itertools
import json
from typing import Any, List, Sequence, Tuple

import numpy as np
import pytest
from pydantic import BaseModel

from useq import (
    Channel,
    GridFromEdges,
    GridRelative,
    MDAEvent,
    MDASequence,
    Position,
    TDurationLoops,
    TIntervalDuration,
    TIntervalLoops,
    ZAboveBelow,
    ZAbsolutePositions,
    ZRangeAround,
    ZRelativePositions,
)
from useq._grid import GridPosition
from useq._pydantic_compat import PYDANTIC2

_T = List[Tuple[Any, Sequence[float]]]

z_as_class: _T = [
    (ZAboveBelow(above=8, below=4, step=2), [-4, -2, 0, 2, 4, 6, 8]),
    (ZAbsolutePositions(absolute=[0, 0.5, 5]), [0, 0.5, 5]),
    (ZRelativePositions(relative=[0, 0.5, 5]), [0, 0.5, 5]),
    (ZRangeAround(range=8, step=1), [-4, -3, -2, -1, 0, 1, 2, 3, 4]),
]
z_as_dict: _T = [
    ({"above": 8, "below": 4, "step": 2}, [-4, -2, 0, 2, 4, 6, 8]),
    ({"range": 8, "step": 1}, [-4, -3, -2, -1, 0, 1, 2, 3, 4]),
    ({"absolute": [0, 0.5, 5]}, [0, 0.5, 5]),
    ({"relative": [0, 0.5, 5]}, [0, 0.5, 5]),
]
z_inputs = z_as_class + z_as_dict

t_as_class: _T = [
    # frame every second for 4 seconds
    (TIntervalDuration(interval=1, duration=4), [0, 1, 2, 3, 4]),
    # 5 frames spanning 8 seconds
    (TDurationLoops(loops=5, duration=8), [0, 2, 4, 6, 8]),
    # 5 frames, taken every 250 ms
    (TIntervalLoops(loops=5, interval=0.25), [0, 0.25, 0.5, 0.75, 1]),
    (
        [
            TIntervalLoops(loops=5, interval=0.25),
            TIntervalDuration(interval=1, duration=4),
        ],
        [0, 0.25, 0.5, 0.75, 1, 2, 3, 4, 5],
    ),
]

t_as_dict: _T = [
    ({"interval": 0.5, "duration": 2}, [0, 0.5, 1, 1.5, 2]),
    ({"loops": 5, "duration": 8}, [0, 2, 4, 6, 8]),
    ({"loops": 5, "interval": 0.25}, [0, 0.25, 0.5, 0.75, 1]),
    (
        [{"loops": 5, "interval": 0.25}, {"interval": 1, "duration": 4}],
        [0, 0.25, 0.50, 0.75, 1, 2, 3, 4, 5],
    ),
    ({"loops": 5, "duration": {"milliseconds": 8}}, [0, 0.002, 0.004, 0.006, 0.008]),
    ({"loops": 5, "duration": {"seconds": 8}}, [0, 2, 4, 6, 8]),
]
t_inputs = t_as_class + t_as_dict

g_inputs = [
    (
        GridRelative(overlap=10, rows=1, columns=2, relative_to="center"),
        [
            GridPosition(x=-0.45, y=0.0, row=0, col=0, is_relative=True),
            GridPosition(x=0.45, y=0.0, row=0, col=1, is_relative=True),
        ],
    ),
    (
        GridRelative(overlap=0, rows=1, columns=2, relative_to="top_left"),
        [
            GridPosition(x=0.0, y=0.0, row=0, col=0, is_relative=True),
            GridPosition(x=1.0, y=0.0, row=0, col=1, is_relative=True),
        ],
    ),
    (
        GridRelative(overlap=(20, 40), rows=2, columns=2),
        [
            GridPosition(x=-0.4, y=0.3, row=0, col=0, is_relative=True),
            GridPosition(x=0.4, y=0.3, row=0, col=1, is_relative=True),
            GridPosition(x=0.4, y=-0.3, row=1, col=1, is_relative=True),
            GridPosition(x=-0.4, y=-0.3, row=1, col=0, is_relative=True),
        ],
    ),
    (
        GridFromEdges(
            overlap=0, top=0, left=0, bottom=20, right=20, fov_height=20, fov_width=20
        ),
        [
            GridPosition(x=0.0, y=20.0, row=0, col=0, is_relative=False),
            GridPosition(x=20.0, y=20.0, row=0, col=1, is_relative=False),
            GridPosition(x=20.0, y=0.0, row=1, col=1, is_relative=False),
            GridPosition(x=0.0, y=0.0, row=1, col=0, is_relative=False),
        ],
    ),
]

all_orders = ["".join(i) for i in itertools.permutations("tpgcz")]

c_inputs = [
    ("DAPI", ("Channel", "DAPI")),
    ({"config": "DAPI"}, ("Channel", "DAPI")),
    ({"config": "DAPI", "group": "Group", "acquire_every": 3}, ("Group", "DAPI")),
    (Channel(config="DAPI"), ("Channel", "DAPI")),
    (Channel(config="DAPI", group="Group"), ("Group", "DAPI")),
]

p_inputs = [
    ([{"x": 0, "y": 1, "z": 2}], (0, 1, 2)),
    ([{"y": 200}], (None, 200, None)),
    ([(100, 200, 300)], (100, 200, 300)),
    (
        [
            {
                "z": 100,
                "sequence": {"z_plan": {"above": 8, "below": 4, "step": 2}},
            }
        ],
        (None, None, 100),
    ),
    ([np.ones(3)], (1, 1, 1)),
    ([(None, 200, None)], (None, 200, None)),
    ([np.ones(2)], (1, 1, None)),
    (np.array([[0, 0, 0], [1, 1, 1]]), (0, 0, 0)),
    (np.array([0, 0]), (0, 0, None)),
    ([Position(x=100, y=200, z=300)], (100, 200, 300)),
]


@pytest.mark.parametrize("zplan, zexpectation", z_inputs)
def test_z_plan(zplan: Any, zexpectation: Sequence[float]) -> None:
    z_plan = MDASequence(z_plan=zplan).z_plan
    assert z_plan and list(z_plan) == zexpectation
    assert z_plan.num_positions() == len(zexpectation)


@pytest.mark.parametrize("gridplan, gridexpectation", g_inputs)
def test_g_plan(gridplan: Any, gridexpectation: Sequence[Any]) -> None:
    g_plan = MDASequence(grid_plan=gridplan).grid_plan
    assert g_plan and list(g_plan) == gridexpectation
    assert g_plan.num_positions() == len(gridexpectation)


@pytest.mark.parametrize("tplan, texpectation", t_inputs)
def test_time_plan(tplan: Any, texpectation: Sequence[float]) -> None:
    time_plan = MDASequence(time_plan=tplan).time_plan
    assert time_plan and list(time_plan) == texpectation
    assert time_plan.num_timepoints() == len(texpectation)


@pytest.mark.parametrize("channel, cexpectation", c_inputs)
def test_channel(channel: Any, cexpectation: Sequence[float]) -> None:
    channel = MDASequence(channels=[channel]).channels[0]
    assert (channel.group, channel.config) == cexpectation


@pytest.mark.parametrize("position, pexpectation", p_inputs)
def test_stage_positions(position: Any, pexpectation: Sequence[float]) -> None:
    position = MDASequence(stage_positions=position).stage_positions[0]
    assert (position.x, position.y, position.z) == pexpectation


def test_axis_order_errors() -> None:
    with pytest.raises(ValueError, match="acquisition order must be a"):
        MDASequence(axis_order=1)
    with pytest.raises(ValueError, match="Duplicate entries found"):
        MDASequence(axis_order="tpgcztpgcz")

    # p after z not ok when z_plan in stage_positions
    with pytest.raises(ValueError, match="'z' cannot precede 'p' in acquisition"):
        MDASequence(
            axis_order="zpc",
            z_plan={"top": 6, "bottom": 0, "step": 1},
            channels=["DAPI"],
            stage_positions=[
                {
                    "x": 0,
                    "y": 0,
                    "z": 0,
                    "sequence": {"z_plan": {"range": 2, "step": 1}},
                }
            ],
        )
    # p before z ok
    MDASequence(
        axis_order="pzc",
        z_plan={"top": 6, "bottom": 0, "step": 1},
        channels=["DAPI"],
        stage_positions=[
            {
                "x": 0,
                "y": 0,
                "z": 0,
                "sequence": {"z_plan": {"range": 2, "step": 1}},
            }
        ],
    )

    # c precedes t not ok if acquire_every > 1 in channels
    with pytest.warns(UserWarning, match="Channels with skipped frames detected"):
        MDASequence(
            axis_order="ct",
            time_plan={"interval": 1, "duration": 10},
            channels=[{"config": "DAPI", "acquire_every": 3}],
        )

    # absolute grid_plan with multiple stage positions

    with pytest.warns(UserWarning, match="Global grid plan will override"):
        MDASequence(
            stage_positions=[(0, 0, 0), (10, 10, 10)],
            grid_plan={"top": 1, "bottom": -1, "left": 0, "right": 0},
        )

    # if grid plan is relative, is ok
    MDASequence(
        stage_positions=[(0, 0, 0), (10, 10, 10)],
        grid_plan={"rows": 2, "columns": 2},
    )

    # if all but one sub-position has a grid plan , is ok
    MDASequence(
        stage_positions=[
            (0, 0, 0),
            {"sequence": {"grid_plan": {"rows": 2, "columns": 2}}},
            {
                "sequence": {
                    "grid_plan": {"top": 1, "bottom": -1, "left": 0, "right": 0}
                }
            },
        ],
        grid_plan={"top": 1, "bottom": -1, "left": 0, "right": 0},
    )

    # multi positions in position sub-sequence
    with pytest.raises(ValueError, match="Currently, a Position sequence cannot"):
        MDASequence(
            stage_positions=[
                {"sequence": {"stage_positions": [(10, 10, 10), (20, 20, 20)]}}
            ]
        )


@pytest.mark.parametrize("tplan, texpectation", t_as_dict[1:3])
@pytest.mark.parametrize("zplan, zexpectation", z_as_dict[:2])
@pytest.mark.parametrize("channel, cexpectation", c_inputs[:3])
@pytest.mark.parametrize("positions, pexpectation", p_inputs[:3])
def test_combinations(
    tplan: Any,
    texpectation: Sequence[float],
    zplan: Any,
    zexpectation: Sequence[float],
    channel: Any,
    cexpectation: Sequence[str],
    positions: Any,
    pexpectation: Sequence[float],
) -> None:
    mda = MDASequence(
        time_plan=tplan, z_plan=zplan, channels=[channel], stage_positions=positions
    )

    assert list(mda.z_plan) == zexpectation
    assert list(mda.time_plan) == texpectation
    assert (mda.channels[0].group, mda.channels[0].config) == cexpectation
    position = mda.stage_positions[0]
    assert (position.x, position.y, position.z) == pexpectation


@pytest.mark.parametrize("cls", [MDASequence, MDAEvent])
def test_schema(cls: BaseModel) -> None:
    if PYDANTIC2:
        schema = cls.model_json_schema()
        assert schema
        assert json.dumps(schema)
    else:
        assert cls.schema()
        assert cls.schema_json()


def test_z_position() -> None:
    mda = MDASequence(axis_order="tpcz", stage_positions=[(222, 1, 10), (111, 1, 20)])
    assert not mda.z_plan
    for event in mda:
        assert event.z_pos


def test_shape_and_axes() -> None:
    mda = MDASequence(
        z_plan=z_as_class[0][0], time_plan=t_as_class[0][0], axis_order="tzp"
    )
    assert mda.shape == (5, 7)
    assert mda.axis_order == "tzp"
    assert mda.used_axes == "tz"
    assert mda.sizes == {"t": 5, "z": 7, "p": 0}

    mda2 = mda.replace(axis_order="zptc")
    assert mda2.shape == (7, 5)
    assert mda2.axis_order == "zptc"
    assert mda2.used_axes == "zt"
    assert mda2.sizes == {"z": 7, "p": 0, "t": 5, "c": 0}

    assert mda2.uid != mda.uid

    with pytest.raises(ValueError):
        mda.replace(axis_order="zptasdfs")


def test_hashable(mda1: MDASequence) -> None:
    assert hash(mda1)
    assert mda1 == mda1
    assert mda1 != 23


def test_mda_str_repr(mda1: MDASequence) -> None:
    assert str(mda1)
    assert repr(mda1)


def test_skip_channel_do_stack_no_zplan() -> None:
    mda = MDASequence(channels=[{"config": "DAPI", "do_stack": False}])
    assert len(list(mda)) == 1


def test_event_action_union() -> None:
    # test that action unions work
    MDAEvent(action={"autofocus_device_name": "Z", "autofocus_motor_offset": 25})


def test_set_fov_deprecation() -> None:
    seq = MDASequence(grid_plan=GridRelative(overlap=10, rows=1, columns=2))
    assert seq.grid_plan
    assert list(seq.grid_plan) == [(-0.45, 0.0, 0, 0, True), (0.45, 0.0, 0, 1, True)]
    with pytest.warns(match="deprecated"):
        seq.set_fov_size((20, 20))
    assert seq.grid_plan.fov_width == 20
    assert list(seq.grid_plan) == [(-9, 0.0, 0, 0, True), (9, 0.0, 0, 1, True)]


def test_keep_shutter_open() -> None:
    # with z as the last axis, the shutter will be left open
    # whenever z is the first index (since there are only 2 z planes)
    mda1 = MDASequence(
        axis_order="tcz",
        channels=["DAPI", "FITC"],
        time_plan=TIntervalLoops(loops=2, interval=0),
        z_plan=ZRangeAround(range=1, step=1),
        keep_shutter_open_across="z",
    )
    assert all(e.keep_shutter_open for e in mda1 if e.index["z"] == 0)

    # with c as the last axis, the shutter will never be left open
    mda2 = MDASequence(
        axis_order="tzc",
        channels=["DAPI", "FITC"],
        time_plan=TIntervalLoops(loops=2, interval=0),
        z_plan=ZRangeAround(range=1, step=1),
        keep_shutter_open_across="z",
    )
    assert not any(e.keep_shutter_open for e in mda2)

    # because t is changing faster than z, the shutter will never be left open
    mda3 = MDASequence(
        axis_order="czt",
        channels=["DAPI", "FITC"],
        time_plan=TIntervalLoops(loops=2, interval=0),
        z_plan=ZRangeAround(range=1, step=1),
        keep_shutter_open_across="z",
    )
    assert not any(e.keep_shutter_open for e in mda3)

    # but, if we include 't' in the keep_shutter_open_across,
    # it will be left open except when it's the last t and last z
    mda4 = MDASequence(
        axis_order="czt",
        channels=["DAPI", "FITC"],
        time_plan=TIntervalLoops(loops=2, interval=0),
        z_plan=ZRangeAround(range=1, step=1),
        keep_shutter_open_across=("z", "t"),
    )
    for event in mda4:
        is_last_zt = bool(event.index["t"] == 1 and event.index["z"] == 1)
        assert event.keep_shutter_open != is_last_zt

    # even though c is the last axis, and comes after g, because the grid happens
    # on a subsequence shutter will be open across the grid for each position
    subseq = MDASequence(grid_plan=GridRelative(rows=2, columns=2))
    mda5 = MDASequence(
        axis_order="pgc",
        channels=["DAPI", "FITC"],
        stage_positions=[Position(sequence=subseq)],
        keep_shutter_open_across="g",
    )
    for event in mda5:
        assert event.keep_shutter_open != (event.index["g"] == 3)



import pytest

from useq import MDASequence


@pytest.fixture
def mda1() -> MDASequence:
    return MDASequence(
        axis_order="tpgcz",
        metadata={"some info": "something"},
        stage_positions=[
            (10, 20),
            {
                "x": 10,
                "y": 20,
                "z": 50,
                "name": "test_name",
                "sequence": MDASequence(
                    z_plan={"above": 10, "below": 0, "step": 1},
                    grid_plan={"rows": 2, "columns": 3},
                ),
            },
        ],
        channels=[
            {"config": "Cy5", "exposure": 50},
            {"config": "FITC", "exposure": 100.0},
            {"config": "DAPI", "do_stack": False, "acquire_every": 3},
        ],
        time_plan=[
            {"interval": 3, "loops": 3},
            {"duration": {"minutes": 40}, "interval": 10},
        ],
        z_plan={"range": 1.0, "step": 0.5},
        grid_plan={"rows": 2, "columns": 1},
        autofocus_plan={
            "autofocus_device_name": "Z",
            "autofocus_motor_offset": 50,
            "axes": ("c",),
        },
    )



from __future__ import annotations

import contextlib
import math
from enum import Enum
from functools import partial
from typing import Any, Callable, Iterator, NamedTuple, Optional, Sequence, Tuple, Union

import numpy as np
from pydantic import Field

from useq._base_model import FrozenModel
from useq._pydantic_compat import FROZEN, PYDANTIC2, field_validator


class RelativeTo(Enum):
    center = "center"
    top_left = "top_left"


class OrderMode(Enum):
    """Different ways of ordering the grid positions."""

    row_wise = "row_wise"
    column_wise = "column_wise"
    row_wise_snake = "row_wise_snake"
    column_wise_snake = "column_wise_snake"
    spiral = "spiral"


def _spiral_indices(
    rows: int, columns: int, center_origin: bool = False
) -> Iterator[Tuple[int, int]]:
    """Return a spiral iterator over a 2D grid.

    Parameters
    ----------
    rows : int
        Number of rows.
    columns : int
        Number of columns.
    center_origin : bool
        If center_origin is True, all indices are centered around (0, 0), and some will
        be negative. Otherwise, the indices are centered around (rows//2, columns//2)

    Yields
    ------
    (x, y) : tuple[int, int]
        Indices of the next element in the spiral.
    """
    # direction: first down and then clockwise (assuming positive Y is down)

    x = y = 0
    if center_origin:  # see docstring
        xshift = yshift = 0
    else:
        xshift = (columns - 1) // 2
        yshift = (rows - 1) // 2
    dx = 0
    dy = -1
    for _ in range(max(columns, rows) ** 2):
        if (-columns / 2 < x <= columns / 2) and (-rows / 2 < y <= rows / 2):
            yield y + yshift, x + xshift
        if x == y or (x < 0 and x == -y) or (x > 0 and x == 1 - y):
            dx, dy = -dy, dx
        x, y = x + dx, y + dy


# function that iterates indices (row, col) in a grid where (0, 0) is the top left
def _rect_indices(
    rows: int, columns: int, snake: bool = False, row_wise: bool = True
) -> Iterator[Tuple[int, int]]:
    """Return a row or column-wise iterator over a 2D grid."""
    c, r = np.meshgrid(np.arange(columns), np.arange(rows))
    if snake:
        if row_wise:
            c[1::2, :] = c[1::2, :][:, ::-1]
        else:
            r[:, 1::2] = r[:, 1::2][::-1, :]
    return zip(r.ravel(), c.ravel()) if row_wise else zip(r.T.ravel(), c.T.ravel())


# used in iter_indices below, to determine the order in which indices are yielded
IndexGenerator = Callable[[int, int], Iterator[Tuple[int, int]]]
_INDEX_GENERATORS: dict[OrderMode, IndexGenerator] = {
    OrderMode.row_wise: partial(_rect_indices, snake=False, row_wise=True),
    OrderMode.column_wise: partial(_rect_indices, snake=False, row_wise=False),
    OrderMode.row_wise_snake: partial(_rect_indices, snake=True, row_wise=True),
    OrderMode.column_wise_snake: partial(_rect_indices, snake=True, row_wise=False),
    OrderMode.spiral: _spiral_indices,
}


class GridPosition(NamedTuple):
    x: float
    y: float
    row: int
    col: int
    is_relative: bool


class _GridPlan(FrozenModel):
    """Base class for all grid plans.

    Attributes
    ----------
    overlap : float | Tuple[float, float]
        Overlap between grid positions in percent. If a single value is provided, it is
        used for both x and y. If a tuple is provided, the first value is used
        for x and the second for y.
    mode : OrderMode
        Define the ways of ordering the grid positions. Options are
        row_wise, column_wise, row_wise_snake, column_wise_snake and spiral.
        By default, row_wise_snake.
    fov_width : Optional[float]
        Width of the field of view in microns.  If not provided, acquisition engines
        should use current width of the FOV based on the current objective and camera.
        Engines MAY override this even if provided.
    fov_height : Optional[float]
        Height of the field of view in microns. If not provided, acquisition engines
        should use current height of the FOV based on the current objective and camera.
        Engines MAY override this even if provided.
    """

    # Overriding FrozenModel to make fov_width and fov_height mutable.
    if PYDANTIC2:
        model_config = {"validate_assignment": True, "frozen": False}
    else:

        class Config:
            validate_assignment = True
            frozen = False

    overlap: Tuple[float, float] = Field((0.0, 0.0), **FROZEN)  # type: ignore
    mode: OrderMode = Field(OrderMode.row_wise_snake, **FROZEN)  # type: ignore
    fov_width: Optional[float] = Field(None)
    fov_height: Optional[float] = Field(None)

    @field_validator("overlap", mode="before")
    def _validate_overlap(cls, v: Any) -> Tuple[float, float]:
        with contextlib.suppress(TypeError, ValueError):
            v = float(v)
        if isinstance(v, float):
            return (v,) * 2
        if isinstance(v, Sequence) and len(v) == 2:
            return float(v[0]), float(v[1])
        raise ValueError(  # pragma: no cover
            "overlap must be a float or a tuple of two floats"
        )

    @property
    def is_relative(self) -> bool:
        return False

    def _offset_x(self, dx: float) -> float:
        raise NotImplementedError

    def _offset_y(self, dy: float) -> float:
        raise NotImplementedError

    def _nrows(self, dy: float) -> int:
        """Return the number of rows, given a grid step size."""
        raise NotImplementedError

    def _ncolumns(self, dx: float) -> int:
        """Return the number of columns, given a grid step size."""
        raise NotImplementedError

    def num_positions(self) -> int:
        """Return the number of individual positions in the grid.

        Note: For GridFromEdges, this will depend on field of view size.  If no
        field of view size is provided, the number of positions will be 1.
        """
        if not self.is_relative and (self.fov_width is None or self.fov_height is None):
            raise ValueError(
                "Retrieving the number of positions in a GridFromEdges plan requires "
                "that the field of view size be set."
            )

        dx, dy = self._step_size(self.fov_width or 1, self.fov_height or 1)
        rows = self._nrows(dy)
        cols = self._ncolumns(dx)
        return rows * cols

    def iter_grid_positions(
        self,
        fov_width: float | None = None,
        fov_height: float | None = None,
        *,
        mode: OrderMode | None = None,
    ) -> Iterator[GridPosition]:
        """Iterate over all grid positions, given a field of view size."""
        _fov_width = fov_width or self.fov_width or 1.0
        _fov_height = fov_height or self.fov_height or 1.0
        mode = self.mode if mode is None else OrderMode(mode)

        dx, dy = self._step_size(_fov_width, _fov_height)
        rows = self._nrows(dy)
        cols = self._ncolumns(dx)
        x0 = self._offset_x(dx)
        y0 = self._offset_y(dy)

        for r, c in _INDEX_GENERATORS[mode](rows, cols):
            yield GridPosition(x0 + c * dx, y0 - r * dy, r, c, self.is_relative)

    def __iter__(self) -> Iterator[GridPosition]:  # type: ignore
        yield from self.iter_grid_positions()

    def _step_size(self, fov_width: float, fov_height: float) -> Tuple[float, float]:
        dx = fov_width - (fov_width * self.overlap[0]) / 100
        dy = fov_height - (fov_height * self.overlap[1]) / 100
        return dx, dy


class GridFromEdges(_GridPlan):
    """Yield absolute stage positions to cover a bounded area.

    The bounded area is defined by top, left, bottom and right edges in
    stage coordinates.

    Attributes
    ----------
    top : float
        Top stage position of the bounding area
    left : float
        Left stage position of the bounding area
    bottom : float
        Bottom stage position of the bounding area
    right : float
        Right stage position of the bounding area
    """

    # everything but fov_width and fov_height is immutable
    top: float = Field(..., **FROZEN)  # type: ignore
    left: float = Field(..., **FROZEN)  # type: ignore
    bottom: float = Field(..., **FROZEN)  # type: ignore
    right: float = Field(..., **FROZEN)  # type: ignore

    def _nrows(self, dy: float) -> int:
        total_height = abs(self.top - self.bottom) + dy
        return math.ceil(total_height / dy)

    def _ncolumns(self, dx: float) -> int:
        total_width = abs(self.right - self.left) + dx
        return math.ceil(total_width / dx)

    def _offset_x(self, dx: float) -> float:
        return min(self.left, self.right)

    def _offset_y(self, dy: float) -> float:
        return max(self.top, self.bottom)


class GridRelative(_GridPlan):
    """Yield relative delta increments to build a grid acquisition.

    Attributes
    ----------
    rows: int
        Number of rows.
    columns: int
        Number of columns.
    relative_to : RelativeTo
        Point in the grid to which the coordinates are relative. If "center", the grid
        is centered around the origin. If "top_left", the grid is positioned such that
        the top left corner is at the origin.
    """

    # everything but fov_width and fov_height is immutable
    rows: int = Field(..., **FROZEN)  # type: ignore
    columns: int = Field(..., **FROZEN)  # type: ignore
    relative_to: RelativeTo = Field(RelativeTo.center, **FROZEN)  # type: ignore

    @property
    def is_relative(self) -> bool:
        return True

    def _nrows(self, dy: float) -> int:
        return self.rows

    def _ncolumns(self, dx: float) -> int:
        return self.columns

    def _offset_x(self, dx: float) -> float:
        return (
            -((self.columns - 1) * dx) / 2
            if self.relative_to == RelativeTo.center
            else 0.0
        )

    def _offset_y(self, dy: float) -> float:
        return (
            ((self.rows - 1) * dy) / 2 if self.relative_to == RelativeTo.center else 0.0
        )


AnyGridPlan = Union[GridFromEdges, GridRelative]



from __future__ import annotations

from contextlib import nullcontext

import numpy as np
import pytest

import useq
from useq._pydantic_compat import PYDANTIC2


def test_from_numpy() -> None:
    useq.ZRelativePositions(relative=np.arange(-1.5, 1.5, 0.5))


DAPI_10 = useq.Channel(config="DAPI", exposure=10)
FITC_30 = useq.Channel(config="DAPI", exposure=30)
TWO_CH_40 = [DAPI_10, FITC_30]
TWO_POS = [(0, 0), (1, 1)]
Z_4 = useq.ZRangeAround(range=3, step=1)
GRID_4 = useq.GridRelative(columns=2, rows=2)
SEQ_40 = useq.MDASequence(channels=TWO_CH_40)

LOOP_1Sx2 = useq.TIntervalLoops(interval=1, loops=2)
SEQ_40_T2 = useq.MDASequence(channels=TWO_CH_40, time_plan=LOOP_1Sx2)


def test_num_pos() -> None:
    # sanity checks to make sure that time_estimation has good inputs
    assert Z_4.num_positions() == 4
    assert GRID_4.num_positions() == 4


def _sizes_str(seq: useq.MDASequence) -> str:
    # key used for test ids
    sizes = "".join([f"{k}{v}" for k, v in seq.sizes.items() if v])
    if seq.channels and any(not ch.do_stack for ch in seq.channels):
        sizes += "_Zskip"
    return sizes


# a list of (sequence, expected total time)
SEQS_NO_T: dict[useq.MDASequence, float] = {
    useq.MDASequence(channels=[DAPI_10]): 0.01,
    SEQ_40: 0.04,
    SEQ_40.replace(stage_positions=[(0, 0, 0)]): 0.04,
    SEQ_40.replace(stage_positions=TWO_POS): 0.08,
    SEQ_40.replace(stage_positions=TWO_POS, z_plan=Z_4): 0.32,
    SEQ_40.replace(stage_positions=TWO_POS, z_plan=Z_4, grid_plan=GRID_4): 1.28,
}


def _duration_exceeded(seq: useq.MDASequence) -> tuple[float, bool]:
    estimate = seq.estimate_duration()
    return estimate.total_duration, estimate.time_interval_exceeded


@pytest.mark.parametrize("seq", SEQS_NO_T, ids=_sizes_str)
def test_time_estimation_without_t(seq: useq.MDASequence) -> None:
    assert _duration_exceeded(seq) == (SEQS_NO_T[seq], False)


SEQS_WITH_T: dict[useq.MDASequence, float | tuple[float, bool]] = {
    # with a simple time plan, the total duration is (ntime - 1) * interval + per_t_time
    # because the first timepoint is acquired immediately
    useq.MDASequence(channels=[DAPI_10], time_plan={"interval": 1, "loops": 2}): 1.01,
    useq.MDASequence(channels=[DAPI_10], time_plan={"interval": 1, "loops": 3}): 2.01,
    useq.MDASequence(channels=TWO_CH_40, time_plan={"interval": 1, "loops": 4}): 3.04,
    SEQ_40_T2: 1.04,  # 1 position is implied
    SEQ_40_T2.replace(stage_positions=[(0, 0, 0)]): 1.04,
    SEQ_40_T2.replace(stage_positions=TWO_POS): 1.08,
    SEQ_40_T2.replace(stage_positions=TWO_POS, z_plan=Z_4): 1.32,
    # HERE, however, a single timepoint takes 1.28s, longer than the time interval
    # so the total time is 1.28s * n_timepoints = 1.28s * 2 = 2.56s
    SEQ_40_T2.replace(stage_positions=TWO_POS, z_plan=Z_4, grid_plan=GRID_4): (
        2.56,
        True,  # time interval exceeded
    ),
}


@pytest.mark.parametrize("seq", SEQS_WITH_T, ids=_sizes_str)
def test_time_estimation_with_t(seq: useq.MDASequence) -> None:
    expect = SEQS_WITH_T[seq]
    if not isinstance(expect, tuple):
        expect = (expect, False)
    assert _duration_exceeded(seq) == expect


DAPI_10_NOZ = DAPI_10.replace(do_stack=False)
SEQS_WITH_Z_SKIPS: dict[useq.MDASequence, float | tuple[float, bool]] = {
    useq.MDASequence(channels=[DAPI_10], z_plan=Z_4): 0.04,
    useq.MDASequence(channels=[DAPI_10_NOZ], z_plan=Z_4): 0.01,
    useq.MDASequence(channels=[FITC_30], z_plan=Z_4): 0.12,
    useq.MDASequence(channels=[DAPI_10_NOZ, FITC_30], z_plan=Z_4): 0.13,  # 0.01 + 0.12
    useq.MDASequence(channels=[DAPI_10, FITC_30], z_plan=Z_4): 0.16,  # 0.04 + 0.12
    # time interval-limited
    useq.MDASequence(
        channels=[DAPI_10_NOZ, FITC_30], z_plan=Z_4, time_plan=LOOP_1Sx2
    ): 1.13,
    useq.MDASequence(
        channels=[DAPI_10_NOZ, FITC_30],
        stage_positions=TWO_POS,
        z_plan=Z_4,
        time_plan=LOOP_1Sx2,
    ): 1.26,  # two positions = 1s interval + 0.13 * 2
    # acquisition-limited
    useq.MDASequence(
        channels=[DAPI_10],
        z_plan=useq.ZRangeAround(range=199, step=1),
        time_plan=LOOP_1Sx2,
    ): (4.0, True),
    useq.MDASequence(
        channels=[DAPI_10_NOZ],  # with NOZ, the time is 0.01s plus a 1sec T interval
        z_plan=useq.ZRangeAround(range=199, step=1),
        time_plan=LOOP_1Sx2,
    ): 1.01,
}


@pytest.mark.parametrize("seq", SEQS_WITH_Z_SKIPS, ids=_sizes_str)
def test_time_estimation_with_z_skips(seq: useq.MDASequence) -> None:
    expect = SEQS_WITH_Z_SKIPS[seq]
    if not isinstance(expect, tuple):
        expect = (expect, False)
    assert _duration_exceeded(seq) == expect


SEQS_WITH_SUBSEQS: dict[useq.MDASequence, float | tuple[float, bool]] = {
    useq.MDASequence(channels=[DAPI_10], stage_positions=TWO_POS): 0.02,
    useq.MDASequence(
        channels=[DAPI_10],
        stage_positions=[
            (0, 0),  # 0.01
            useq.Position(x=1, sequence=useq.MDASequence(grid_plan=GRID_4)),  # 0.04
        ],
    ): 0.05,  # 0.01 + 0.04
}


@pytest.mark.parametrize("seq", SEQS_WITH_SUBSEQS, ids=_sizes_str)
def test_time_estimation_with_position_seqs(seq: useq.MDASequence) -> None:
    expect = SEQS_WITH_SUBSEQS[seq]
    if not isinstance(expect, tuple):
        expect = (expect, False)
    assert _duration_exceeded(seq) == expect


def test_pydantic_compat(mda1: useq.MDASequence) -> None:
    # testing names of deprecated methods and cross-compatible API
    assert mda1.json()
    assert mda1.model_dump_json()

    assert mda1.model_dump()
    ctx = pytest.warns(DeprecationWarning) if PYDANTIC2 else nullcontext()
    with ctx:  # type: ignore
        assert mda1.dict()



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Dict, Iterator, Optional, Sequence, Tuple
from uuid import UUID, uuid4
from warnings import warn

import numpy as np
from pydantic import Field, PrivateAttr

from useq._base_model import UseqModel
from useq._channel import Channel
from useq._grid import AnyGridPlan, GridPosition  # noqa: TCH001
from useq._hardware_autofocus import AnyAutofocusPlan, AxesBasedAF
from useq._iter_sequence import iter_sequence
from useq._position import Position
from useq._pydantic_compat import (
    field_validator,
    model_construct,
    model_dump,
    model_validator,
    pydantic_1_style_root_dict,
)
from useq._time import AnyTimePlan  # noqa: TCH001
from useq._utils import AXES, Axis, TimeEstimate, estimate_sequence_duration
from useq._z import AnyZPlan  # noqa: TCH001

if TYPE_CHECKING:
    from useq._mda_event import MDAEvent
    from useq.pycromanager import PycroManagerEvent


class MDASequence(UseqModel):
    """A sequence of MDA (Multi-Dimensional Acquisition) events.

    This is the core object in the `useq` library, and is used define a sequence of
    events to be run on a microscope. It object may be constructed manually, or from
    file (e.g. json or yaml).

    The object itself acts as an iterator for [`useq.MDAEvent`][] objects:

    Attributes
    ----------
    metadata : dict
        A dictionary of user metadata to be stored with the sequence.
    axis_order : str
        The order of the axes in the sequence. Must be a permutation of `"tpgcz"`. The
        default is `"tpgcz"`.
    stage_positions : tuple[Position, ...]
        The stage positions to visit. (each with `x`, `y`, `z`, `name`, and `sequence`,
        all of which are optional).
    grid_plan : GridFromEdges | GridRelative | None
        The grid plan to follow. One of `GridFromEdges`, `GridRelative` or `None`.
    channels : tuple[Channel, ...]
        The channels to acquire. see `Channel`.
    time_plan : MultiPhaseTimePlan | TIntervalDuration | TIntervalLoops \
        | TDurationLoops | None
        The time plan to follow. One of `TIntervalDuration`, `TIntervalLoops`,
        `TDurationLoops`, `MultiPhaseTimePlan`, or `None`
    z_plan : ZTopBottom | ZRangeAround | ZAboveBelow | ZRelativePositions | \
        ZAbsolutePositions | None
        The z plan to follow. One of `ZTopBottom`, `ZRangeAround`, `ZAboveBelow`,
        `ZRelativePositions`, `ZAbsolutePositions`, or `None`.
    uid : UUID
        A read-only unique identifier (uuid version 4) for the sequence. This will be
        generated, do not set.
    autofocus_plan : AxesBasedAF | None
        The hardware autofocus plan to follow. One of `AxesBasedAF` or `None`.
    keep_shutter_open_across : tuple[str, ...]
        A tuple of axes `str` across which the illumination shutter should be kept open.
        Resulting events will have `keep_shutter_open` set to `True` if and only if
        ALL axes whose indices are changing are in this tuple. For example, if
        `keep_shutter_open_across=('z',)`, then the shutter would be kept open between
        events axes {'t': 0, 'z: 0} and {'t': 0, 'z': 1}, but not between
        {'t': 0, 'z': 0} and {'t': 1, 'z': 0}.

    Examples
    --------
    >>> from useq import MDASequence, Position, Channel, TIntervalDuration
    >>> seq = MDASequence(
    ...     time_plan={"interval": 0.1, "loops": 2},
    ...     stage_positions=[(1, 1, 1)],
    ...     grid_plan={"rows": 2, "cols": 2},
    ...     z_plan={"range": 3, "step": 1},
    ...     channels=[{"config": "DAPI", "exposure": 1}]
    ... )
    >>> print(seq)
    Multi-Dimensional Acquisition ▶ nt: 2, np: 1, nc: 1, nz: 4, ng: 4

    >>> for event in seq:
    ...     print(event)

    >>> print(seq.yaml())
    channels:
    - config: DAPI
      exposure: 1.0
    grid_plan:
      columns: 2
      rows: 2
    stage_positions:
    - x: 1.0
      y: 1.0
      z: 1.0
    time_plan:
      interval: '0:00:00.100000'
      loops: 2
    z_plan:
      range: 3.0
      step: 1.0
    """

    metadata: Dict[str, Any] = Field(default_factory=dict)
    axis_order: str = "".join(AXES)
    stage_positions: Tuple[Position, ...] = Field(default_factory=tuple)
    grid_plan: Optional[AnyGridPlan] = None
    channels: Tuple[Channel, ...] = Field(default_factory=tuple)
    time_plan: Optional[AnyTimePlan] = None
    z_plan: Optional[AnyZPlan] = None
    autofocus_plan: Optional[AnyAutofocusPlan] = None
    keep_shutter_open_across: Tuple[str, ...] = Field(default_factory=tuple)

    _uid: UUID = PrivateAttr(default_factory=uuid4)
    _sizes: Optional[Dict[str, int]] = PrivateAttr(default=None)

    @property
    def uid(self) -> UUID:
        """A unique identifier for this sequence."""
        return self._uid

    def set_fov_size(self, fov_size: Tuple[float, float]) -> None:
        """Set the field of view size.

        FOV is used to calculate the number of positions in a grid plan.

        !!! warning "DEPRECATED"

            Set `fov_width` and `fov_height` directly on the `grid_plan` instead.
        """
        warn(
            "set_fov_size is deprecated and will be removed. Please directly mutate "
            "fov_width and fov_height on the sequence.grid_plan instead. (They are "
            "mutable.)",
            DeprecationWarning,
            stacklevel=2,
        )
        if self.grid_plan:
            self.grid_plan.fov_width = fov_size[0]
            self.grid_plan.fov_height = fov_size[1]

    def __hash__(self) -> int:
        return hash(self.uid)

    @field_validator("z_plan", mode="before")
    def _validate_zplan(cls, v: Any) -> Optional[dict]:
        return v or None

    @field_validator("keep_shutter_open_across", mode="before")
    def _validate_keep_shutter_open_across(cls, v: tuple[str, ...]) -> tuple[str, ...]:
        try:
            v = tuple(v)
        except (TypeError, ValueError):  # pragma: no cover
            raise ValueError(
                f"keep_shutter_open_across must be string or a sequence of strings, "
                f"got {type(v)}"
            ) from None
        return v

    @field_validator("channels", mode="before")
    def _validate_channels(cls, value: Any) -> Tuple[Channel, ...]:
        if isinstance(value, str) or not isinstance(
            value, Sequence
        ):  # pragma: no cover
            raise ValueError(f"channels must be a sequence, got {type(value)}")
        channels = []
        for v in value:
            if isinstance(v, Channel):
                channels.append(v)
            elif isinstance(v, str):
                channels.append(model_construct(Channel, config=v))
            elif isinstance(v, dict):
                channels.append(Channel(**v))
            else:  # pragma: no cover
                raise ValueError(f"Invalid Channel argument: {value!r}")
        return tuple(channels)

    @field_validator("stage_positions", mode="before")
    def _validate_stage_positions(cls, value: Any) -> Tuple[Position, ...]:
        if isinstance(value, np.ndarray):
            if value.ndim == 1:
                value = [value]
            elif value.ndim == 2:
                value = list(value)
        if not isinstance(value, Sequence):  # pragma: no cover
            raise ValueError(f"stage_positions must be a sequence, got {type(value)}")

        positions = []
        for v in value:
            if isinstance(v, Position):
                positions.append(v)
            elif isinstance(v, dict):
                positions.append(Position(**v))
            elif isinstance(v, (np.ndarray, tuple)):
                x, *v = v
                y, *v = v or (None,)
                z = v[0] if v else None
                positions.append(Position(x=x, y=y, z=z))
            else:  # pragma: no cover
                raise ValueError(f"Cannot coerce {v!r} to Position")
        return tuple(positions)

    @field_validator("time_plan", mode="before")
    def _validate_time_plan(cls, v: Any) -> Optional[dict]:
        return {"phases": v} if isinstance(v, (tuple, list)) else v or None

    @field_validator("axis_order", mode="before")
    def _validate_axis_order(cls, v: Any) -> str:
        if not isinstance(v, str):
            raise ValueError(f"acquisition order must be a string, got {type(v)}")
        order = v.lower()
        extra = {x for x in order if x not in AXES}
        if extra:
            raise ValueError(
                f"Can only iterate over axes: {AXES!r}. Got extra: {extra}"
            )
        if len(set(order)) < len(order):
            raise ValueError(f"Duplicate entries found in acquisition order: {order}")

        return order

    @model_validator(mode="after")
    @classmethod
    def _validate_mda(cls, values: Any) -> Any:
        # this strange bit here is to deal with the fact that in pydantic1
        # root_validator after returned a dict of {field_name -> validated_value}
        # but in pydantic2 it returns the complete validated model instance
        _values = pydantic_1_style_root_dict(cls, values)

        if "axis_order" in _values:
            cls._check_order(
                _values["axis_order"],
                z_plan=_values.get("z_plan"),
                stage_positions=_values.get("stage_positions", ()),
                channels=_values.get("channels", ()),
                grid_plan=_values.get("grid_plan"),
                autofocus_plan=_values.get("autofocus_plan"),
            )
        if "stage_positions" in _values:
            for p in _values["stage_positions"]:
                if hasattr(p, "sequence") and getattr(
                    p.sequence, "keep_shutter_open_across", None
                ):  # pragma: no cover
                    raise ValueError(
                        "keep_shutter_open_across cannot currently be set on a "
                        "Position sequence"
                    )
        return values

    def __eq__(self, other: Any) -> bool:
        """Return `True` if two `MDASequences` are equal (uid is excluded)."""
        if isinstance(other, MDASequence):
            return bool(
                model_dump(self, exclude={"uid"}) == model_dump(other, exclude={"uid"})
            )
        else:
            return False

    @staticmethod
    def _check_order(
        order: str,
        z_plan: Optional[AnyZPlan] = None,
        stage_positions: Sequence[Position] = (),
        channels: Sequence[Channel] = (),
        grid_plan: Optional[AnyGridPlan] = None,
        autofocus_plan: Optional[AnyAutofocusPlan] = None,
    ) -> None:
        if (
            Axis.Z in order
            and Axis.POSITION in order
            and order.index(Axis.Z) < order.index(Axis.POSITION)
            and z_plan
            and any(
                p.sequence.z_plan for p in stage_positions if p.sequence is not None
            )
        ):
            raise ValueError(
                f"{Axis.Z!r} cannot precede {Axis.POSITION!r} in acquisition order if "
                "any position specifies a z_plan"
            )

        if (
            Axis.CHANNEL in order
            and Axis.TIME in order
            and any(c.acquire_every > 1 for c in channels)
            and order.index(Axis.CHANNEL) < order.index(Axis.TIME)
        ):
            warn(
                f"Channels with skipped frames detected, but {Axis.CHANNEL!r} precedes "
                "{TIME!r} in the acquisition order: may not yield intended results.",
                stacklevel=2,
            )

        if (
            Axis.GRID in order
            and Axis.POSITION in order
            and grid_plan
            and not grid_plan.is_relative
            and len(stage_positions) > 1
        ):
            sub_position_grid_plans = [
                p
                for p in stage_positions
                if p.sequence is not None and p.sequence.grid_plan
            ]
            if len(stage_positions) - len(sub_position_grid_plans) > 1:
                warn(
                    "Global grid plan will override sub-position grid plans.",
                    stacklevel=2,
                )

        if (
            Axis.POSITION in order
            and stage_positions
            and any(
                p.sequence.stage_positions
                for p in stage_positions
                if p.sequence is not None
            )
        ):
            raise ValueError(
                "Currently, a Position sequence cannot have multiple stage positions."
            )

        # Cannot use autofocus plan with absolute z_plan
        if Axis.Z in order and z_plan and not z_plan.is_relative:
            err = "Absolute Z positions cannot be used with autofocus plan."
            if isinstance(autofocus_plan, AxesBasedAF):
                raise ValueError(err)
            for p in stage_positions:
                if p.sequence is not None and p.sequence.autofocus_plan:
                    raise ValueError(err)

    def __str__(self) -> str:
        return "Multi-Dimensional Acquisition ▶ " + ", ".join(self.sizes)

    @property
    def shape(self) -> Tuple[int, ...]:
        """Return the shape of this sequence.

        !!! note
            This doesn't account for jagged arrays, like channels that exclude z
            stacks or skip timepoints.
        """
        return tuple(s for s in self.sizes.values() if s)

    @property
    def sizes(self) -> Dict[str, int]:
        """Mapping of axis name to size of that axis."""
        if self._sizes is None:
            self._sizes = {k: len(list(self.iter_axis(k))) for k in self.axis_order}
        return self._sizes

    @property
    def used_axes(self) -> str:
        """Single letter string of axes used in this sequence, e.g. `ztc`."""
        return "".join(k for k in self.axis_order if self.sizes[k])

    def iter_axis(
        self, axis: str
    ) -> Iterator[Position | Channel | float | GridPosition]:
        """Iterate over the positions or items of a given axis."""
        plan = {
            Axis.TIME: self.time_plan,
            Axis.POSITION: self.stage_positions,
            Axis.Z: self.z_plan,
            Axis.CHANNEL: self.channels,
            Axis.GRID: self.grid_plan,
        }[axis]
        if plan:
            yield from plan

    def __iter__(self) -> Iterator[MDAEvent]:  # type: ignore [override]
        """Same as `iter_events`. Supports `for event in sequence: ...` syntax."""
        yield from self.iter_events()

    def iter_events(self) -> Iterator[MDAEvent]:
        """Iterate over all events in the MDA sequence.

        See source of [useq._mda_sequence.iter_sequence][] for details on how
        events are constructed and yielded.

        Yields
        ------
        MDAEvent
            Each event in the MDA sequence.
        """
        return iter_sequence(self)

    def to_pycromanager(self) -> list[PycroManagerEvent]:
        warnings.warn(
            "useq.MDASequence.to_pycromanager() is deprecated and will be removed in a "
            "future version. Useq useq.pycromanager.to_pycromanager(seq) instead.",
            FutureWarning,
            stacklevel=2,
        )
        from useq.pycromanager import to_pycromanager

        return to_pycromanager(self)

    def estimate_duration(self) -> TimeEstimate:
        """Estimate duration and other timing issues of an MDASequence.

        Notable mis-estimations may include:
        - when the time interval is shorter than the time it takes to acquire the data
        and any of the channels have `acquire_every` > 1
        - when channel exposure times are omitted. In this case, we assume 1ms exposure.

        Returns
        -------
        TimeEstimate
            A named 3-tuple with the following fields:
            - total_duration: float
                Estimated total duration of the experiment, in seconds.
            - per_t_duration: float
                Estimated duration of a single timepoint, in seconds.
            - time_interval_exceeded: bool
                Whether the time interval between timepoints is shorter than the time it
                takes to acquire the data
        """
        return estimate_sequence_duration(self)



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from useq._base_model import FrozenModel

if TYPE_CHECKING:
    from useq import MDASequence


class Position(FrozenModel):
    """Define a position in 3D space.

    Any of the attributes can be `None` to indicate that the position is not
    defined. This is useful for defining a position relative to the current
    position.

    Attributes
    ----------
    x : float | None
        X position in microns.
    y : float | None
        Y position in microns.
    z : float | None
        Z position in microns.
    name : str | None
        Optional name for the position.
    sequence : MDASequence | None
        Optional MDASequence relative this position.
    """

    # if None, implies 'do not move this axis'
    x: Optional[float] = None
    y: Optional[float] = None
    z: Optional[float] = None
    name: Optional[str] = None
    sequence: Optional[MDASequence] = None



import pytest

from useq import GridFromEdges, GridRelative
from useq._grid import OrderMode, _rect_indices, _spiral_indices

EXPECT = {
    (True, False): [(0, 0), (0, 1), (1, 0), (1, 1), (2, 0), (2, 1)],
    (True, True): [(0, 0), (0, 1), (1, 1), (1, 0), (2, 0), (2, 1)],
    (False, False): [(0, 0), (1, 0), (2, 0), (0, 1), (1, 1), (2, 1)],
    (False, True): [(0, 0), (1, 0), (2, 0), (2, 1), (1, 1), (0, 1)],
}


@pytest.mark.parametrize("row_wise", [True, False], ids=["row_wise", "col_wise"])
@pytest.mark.parametrize("snake", [True, False], ids=["snake", "normal"])
def test_grid_indices(row_wise: bool, snake: bool) -> None:
    indices = _rect_indices(3, 2, snake=snake, row_wise=row_wise)
    assert list(indices) == EXPECT[(row_wise, snake)]


def test_spiral_indices() -> None:
    assert list(_spiral_indices(2, 3)) == [
        (0, 1),
        (0, 2),
        (1, 2),
        (1, 1),
        (1, 0),
        (0, 0),
    ]
    assert list(_spiral_indices(2, 3, center_origin=True)) == [
        (0, 0),
        (0, 1),
        (1, 1),
        (1, 0),
        (1, -1),
        (0, -1),
    ]


def test_position_equality():
    """Order of grid positions should only change the order in which they are yielded"""
    t1 = GridRelative(rows=3, columns=3, mode=OrderMode.spiral)
    spiral_pos = set(t1.iter_grid_positions(1, 1))

    t2 = GridRelative(rows=3, columns=3, mode=OrderMode.row_wise)
    row_pos = set(t2.iter_grid_positions(1, 1))

    t3 = GridRelative(rows=3, columns=3, mode="row_wise_snake")
    snake_row_pos = set(t3.iter_grid_positions(1, 1))

    t4 = GridRelative(rows=3, columns=3, mode=OrderMode.column_wise)
    col_pos = set(t4.iter_grid_positions(1, 1))

    t5 = GridRelative(rows=3, columns=3, mode=OrderMode.column_wise_snake)
    snake_col_pos = set(t5.iter_grid_positions(1, 1))

    assert spiral_pos == row_pos == snake_row_pos == col_pos == snake_col_pos


def test_grid_type():
    g1 = GridRelative(rows=2, columns=3)
    assert [(g.x, g.y) for g in g1.iter_grid_positions(1, 1)] == [
        (-1.0, 0.5),
        (0.0, 0.5),
        (1.0, 0.5),
        (1.0, -0.5),
        (0.0, -0.5),
        (-1.0, -0.5),
    ]
    g2 = GridFromEdges(top=1, left=-1, bottom=-1, right=2)
    assert [(g.x, g.y) for g in g2.iter_grid_positions(1, 1)] == [
        (-1.0, 1.0),
        (0.0, 1.0),
        (1.0, 1.0),
        (2.0, 1.0),
        (2.0, 0.0),
        (1.0, 0.0),
        (0.0, 0.0),
        (-1.0, 0.0),
        (-1.0, -1.0),
        (0.0, -1.0),
        (1.0, -1.0),
        (2.0, -1.0),
    ]


def test_num_position_error() -> None:
    with pytest.raises(ValueError, match="the field of view size be set"):
        GridFromEdges(top=1, left=-1, bottom=-1, right=2).num_positions()



import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generator,
    Iterator,
    Sequence,
    Union,
)

from pydantic import Field

from useq._base_model import FrozenModel
from useq._utils import parse_duration

if TYPE_CHECKING:
    from pydantic_core import CoreSchema, core_schema


# FIXME: please!!
# This is a gross amalgamation of fixes that tries to work with both pydantic1 and 2
class timedelta(datetime.timedelta):
    @classmethod
    def __get_validators__(cls) -> Generator[Callable[..., Any], None, None]:
        yield cls.validate

    @classmethod
    def validate(cls, v: Any) -> datetime.timedelta:
        return datetime.timedelta(**v) if isinstance(v, dict) else parse_duration(v)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: Any
    ) -> "CoreSchema":
        from pydantic_core.core_schema import (
            no_info_plain_validator_function,
            plain_serializer_function_ser_schema,
        )

        serializer = plain_serializer_function_ser_schema(
            cls._serialize, when_used="json"
        )

        return no_info_plain_validator_function(cls.validate, serialization=serializer)

    @classmethod
    def _serialize(cls, v: datetime.timedelta) -> float:
        return v.total_seconds()

    @classmethod
    def __get_pydantic_json_schema__(
        cls, core_schema: "core_schema.CoreSchema", handler: Any
    ) -> Dict[str, Any]:
        return {"type": "number", "format": "float"}


class TimePlan(FrozenModel):
    # TODO: probably needs to be implemented by engine
    prioritize_duration: bool = False  # or prioritize num frames

    def __iter__(self) -> Iterator[float]:  # type: ignore
        for td in self.deltas():
            yield td.total_seconds()

    def num_timepoints(self) -> int:
        return self.loops  # type: ignore  # TODO

    def deltas(self) -> Iterator[datetime.timedelta]:
        current = timedelta(0)
        for _ in range(self.loops):  # type: ignore  # TODO
            yield current
            current += self.interval  # type: ignore  # TODO


class TIntervalLoops(TimePlan):
    """Define temporal sequence using interval and number of loops.

    Attributes
    ----------
    interval : str | timedelta | float
        Time between frames. Scalars are interpreted as seconds.
        Strings are parsed according to ISO 8601.
    loops : int
        Number of frames.
    prioritize_duration : bool
        If `True`, instructs engine to prioritize duration over number of frames in case
        of conflict. By default, `False`.
    """

    interval: timedelta
    loops: int = Field(..., gt=0)


class TDurationLoops(TimePlan):
    """Define temporal sequence using duration and number of loops.

    Attributes
    ----------
    duration : str | timedelta
        Total duration of sequence. Scalars are interpreted as seconds.
        Strings are parsed according to ISO 8601.
    loops : int
        Number of frames.
    prioritize_duration : bool
        If `True`, instructs engine to prioritize duration over number of frames in case
        of conflict. By default, `False`.
    """

    duration: timedelta
    loops: int = Field(..., gt=0)

    @property
    def interval(self) -> datetime.timedelta:
        # -1 makes it so that the last loop will *occur* at duration, not *finish*
        return self.duration / (self.loops - 1)


class TIntervalDuration(TimePlan):
    """Define temporal sequence using interval and duration.

    Attributes
    ----------
    interval : str | timedelta
        Time between frames. Scalars are interpreted as seconds.
        Strings are parsed according to ISO 8601.
    duration : str | timedelta
        Total duration of sequence.
    prioritize_duration : bool
        If `True`, instructs engine to prioritize duration over number of frames in case
        of conflict. By default, `True`.
    """

    interval: timedelta
    duration: timedelta
    prioritize_duration: bool = True

    @property
    def loops(self) -> int:
        return self.duration // self.interval + 1


SinglePhaseTimePlan = Union[TIntervalDuration, TIntervalLoops, TDurationLoops]


class MultiPhaseTimePlan(TimePlan):
    """Time sequence composed of multiple phases.

    Attributes
    ----------
    phases : Sequence[TIntervalDuration | TIntervalLoops | TDurationLoops]
        Sequence of time plans.
    """

    phases: Sequence[SinglePhaseTimePlan]

    def deltas(self) -> Iterator[datetime.timedelta]:
        accum = datetime.timedelta(0)
        yield accum
        for phase in self.phases:
            for i, td in enumerate(phase.deltas()):
                # skip the first timepoint of later phases
                if i == 0 and td == datetime.timedelta(0):
                    continue
                yield td + accum
            accum += td

    def num_timepoints(self) -> int:
        # TODO: is this correct?
        return sum(phase.loops for phase in self.phases) - 1


AnyTimePlan = Union[MultiPhaseTimePlan, SinglePhaseTimePlan]



from useq._actions import AcquireImage, Action, HardwareAutofocus
from useq._channel import Channel
from useq._grid import AnyGridPlan, GridFromEdges, GridRelative
from useq._hardware_autofocus import AnyAutofocusPlan, AutoFocusPlan, AxesBasedAF
from useq._mda_event import MDAEvent, PropertyTuple
from useq._mda_sequence import MDASequence
from useq._position import Position
from useq._time import (
    MultiPhaseTimePlan,
    TDurationLoops,
    TIntervalDuration,
    TIntervalLoops,
)
from useq._z import (
    ZAboveBelow,
    ZAbsolutePositions,
    ZRangeAround,
    ZRelativePositions,
    ZTopBottom,
)

__all__ = [
    "AcquireImage",
    "Action",
    "AnyAutofocusPlan",
    "AnyGridPlan",
    "AutoFocusPlan",
    "AxesBasedAF",
    "Channel",
    "GridFromEdges",
    "GridRelative",
    "HardwareAutofocus",
    "MDAEvent",
    "MDASequence",
    "MultiPhaseTimePlan",
    "Position",
    "PropertyTuple",
    "TDurationLoops",
    "TIntervalDuration",
    "TIntervalLoops",
    "ZAboveBelow",
    "ZAbsolutePositions",
    "ZRangeAround",
    "ZRelativePositions",
    "ZTopBottom",
]

from useq._pydantic_compat import model_rebuild

model_rebuild(MDAEvent, MDASequence=MDASequence)
model_rebuild(Position, MDASequence=MDASequence)
del model_rebuild



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, TypeVar

import pydantic.version
from pydantic import BaseModel

if TYPE_CHECKING:
    from pydantic.fields import FieldInfo


PYDANTIC2 = pydantic.version.VERSION.startswith("2")
BM = TypeVar("BM", bound=BaseModel)

__all__ = [
    "PYDANTIC2",
    "model_validator",
    "field_validator",
    "field_serializer",
    "model_serializer",
    "model_fields",
]


if PYDANTIC2:
    from pydantic import (
        field_serializer,
        functional_validators,
        model_serializer,
        model_validator,
    )

    def model_fields(obj: BaseModel | type[BaseModel]) -> dict[str, FieldInfo]:
        return obj.model_fields

    def field_validator(*args: Any, **kwargs: Any) -> Callable[[Callable], Callable]:
        kwargs.pop("always", None)
        kwargs.pop("allow_reuse", None)
        return functional_validators.field_validator(*args, **kwargs)

    def model_dump(obj: BaseModel, **kwargs: Any) -> dict[str, Any]:
        return obj.model_dump(**kwargs)

    def model_dump_json(obj: BaseModel, **kwargs: Any) -> str:
        return obj.model_dump_json(**kwargs)

    def model_rebuild(obj: type[BaseModel], **kwargs: Any) -> bool | None:
        return obj.model_rebuild(_types_namespace=kwargs)

    def model_construct(obj: type[BM], **kwargs: Any) -> BM:
        return obj.model_construct(**kwargs)

    def model_copy(obj: BM, **kwargs: Any) -> BM:
        return obj.model_copy(**kwargs)

    def pydantic_1_style_root_dict(cls: type[BM], values: BM) -> dict:
        # deal with the fact that in pydantic1
        # root_validator after returned a dict of {field_name -> validated_value}
        # but in pydantic2 it returns the complete validated model instance
        return {k: getattr(values, k) for k in cls.model_fields}

    FROZEN = {"frozen": True}

elif not TYPE_CHECKING:
    from pydantic import root_validator, validator

    def model_fields(
        obj: BaseModel | type[BaseModel],
    ) -> dict[str, Any]:
        return obj.__fields__

    def model_validator(**kwargs: Any) -> Callable[[Callable], Callable]:
        if kwargs.pop("mode", None) == "before":
            kwargs["pre"] = True  # pragma: no cover
        return root_validator(**kwargs)

    def field_validator(*fields: str, **kwargs: Any) -> Callable[[Callable], Callable]:
        if kwargs.pop("mode", None) == "before":
            kwargs["pre"] = True
        return validator(*fields, **kwargs)

    def model_dump(obj: BaseModel, **kwargs: Any) -> dict[str, Any]:
        return obj.dict(**kwargs)

    def model_copy(obj: BM, **kwargs: Any) -> BM:
        return obj.copy(**kwargs)

    def model_dump_json(obj: BaseModel, **kwargs: Any) -> str:
        return obj.json(**kwargs)

    def model_rebuild(obj: type[BM], **kwargs: Any) -> None:
        obj.update_forward_refs(**kwargs)

    def model_construct(obj: type[BM], **kwargs: Any) -> BM:
        return obj.construct(**kwargs)

    def pydantic_1_style_root_dict(cls: type[BM], values: dict) -> dict:
        return values

    def model_serializer(**kwargs):
        return lambda f: f  # pragma: no cover

    def field_serializer(*args, **kwargs):
        return lambda f: f  # pragma: no cover

    FROZEN = {"allow_mutation": False}



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from useq import MDASequence


def test_pycromanager_events(mda1: MDASequence) -> None:
    # after deprecation is removed, use this import directly
    # from useq.pycromanager import to_pycromanager

    with pytest.warns(FutureWarning, match="deprecated"):
        events = mda1.to_pycromanager()
        assert isinstance(events, list)
        assert isinstance(events[0], dict)



from typing import Any, Optional, Tuple

from pydantic import PrivateAttr

from useq._actions import HardwareAutofocus
from useq._base_model import FrozenModel
from useq._mda_event import MDAEvent
from useq._pydantic_compat import model_copy


class AutoFocusPlan(FrozenModel):
    """Base class for hardware autofocus plans.

    Attributes
    ----------
    autofocus_device_name : str
        Name of the hardware autofocus z device.
    autofocus_motor_offset : float | None
        Before autofocus is performed, the autofocus motor should be moved to this
        offset.
    """

    autofocus_device_name: str
    autofocus_motor_offset: Optional[float] = None

    def as_action(self) -> HardwareAutofocus:
        """Return a [`useq.HardwareAutofocus`][] for this autofocus plan."""
        return HardwareAutofocus(
            autofocus_device_name=self.autofocus_device_name,
            autofocus_motor_offset=self.autofocus_motor_offset,
        )

    def event(self, event: MDAEvent) -> Optional[MDAEvent]:
        """Return an autofocus [`useq.MDAEvent`][] if autofocus should be performed.

        The z position of the new [`useq.MDAEvent`][] is also updated if a relative
        zplan is provided since autofocus shuld be performed on the home z stack
        position.
        """
        if not self.should_autofocus(event):
            return None

        updates: dict[str, Any] = {"action": self.as_action()}
        if event.z_pos is not None and event.sequence is not None:
            zplan = event.sequence.z_plan
            if zplan and zplan.is_relative and "z" in event.index:
                updates["z_pos"] = event.z_pos - list(zplan)[event.index["z"]]

        return model_copy(event, update=updates)

    def should_autofocus(self, event: MDAEvent) -> bool:
        """Method that must be implemented by a subclass.

        Should return True if autofocus should be performed (see
        [`useq.AxesBasedAF`][]).
        """
        raise NotImplementedError("should_autofocus() must be implemented by subclass.")


class AxesBasedAF(AutoFocusPlan):
    """Autofocus plan that performs autofocus when any of the specified axes change.

    Attributes
    ----------
    axes : Tuple[str, ...]
        Tuple of axis label to use for hardware autofocus.  At every event in which
        *any* axis in this tuple is change, autofocus will be performed.  For example,
        if `axes` is `('p',)` then autofocus will be performed every time the `p` axis
        is change, (in other words: every time the position is changed.).
    """

    axes: Tuple[str, ...]
    _previous: dict = PrivateAttr(default_factory=dict)

    def should_autofocus(self, event: MDAEvent) -> bool:
        """Return `True` if autofocus should be performed at this event.

        Will return `True` if any of the axes specified in `axes` have changed from the
        previous event.
        """
        self._previous, previous = dict(event.index), self._previous
        return any(
            axis in self.axes and previous.get(axis) != index
            for axis, index in event.index.items()
        )


AnyAutofocusPlan = AxesBasedAF



from __future__ import annotations

from typing import TYPE_CHECKING, overload

from useq import MDAEvent, MDASequence
from useq._utils import Axis

if TYPE_CHECKING:
    from typing_extensions import Literal, Required, TypedDict

    class PycroManagerAxes(TypedDict, total=False):
        """Axes dict used by pycro-manager."""

        time: int
        position: int
        z: int
        channel: str
        row: int
        column: int

    class PycroManagerEvent(TypedDict, total=False):
        """Dict in the format expected by pycro-manager."""

        axes: Required[PycroManagerAxes]
        x: float
        y: float
        z: float
        min_start_time: float
        config_group: list[str]  # [group, config]
        exposure: float
        properties: list[list[str]]  # [[device, property, value], ...]
        keep_shutter_open: bool

    PycroAxis = Literal["time", "position", "z", "channel", "row", "column"]
    PycroKey = Literal["x", "y", "z", "exposure", "min_start_time", "keep_shutter_open"]


@overload
def to_pycromanager(obj: MDAEvent) -> PycroManagerEvent:
    ...


@overload
def to_pycromanager(obj: MDASequence) -> list[PycroManagerEvent]:
    ...


def to_pycromanager(
    obj: MDAEvent | MDASequence,
) -> PycroManagerEvent | list[PycroManagerEvent]:
    """Convert a [`useq.MDAEvent`][] or [`useq.MDASequence`][] to pycro-manager form.

    Parameters
    ----------
    obj : MDAEvent | MDASequence
        The event or sequence to convert.

    Returns
    -------
    PycroManagerEvent | list[PycroManagerEvent]
        If `obj` is an [`useq.MDAEvent`][], returns a single pycro-manager event dict.
        Otherwise, returns a list of pycro-manager event dicts.
    """
    if isinstance(obj, MDAEvent):
        return _event_to_pycromanager(obj)
    elif isinstance(obj, MDASequence):
        return [_event_to_pycromanager(event) for event in obj]
    raise TypeError(f"invalid argument: {obj!r}. Must be MDAEvent or MDASequence.")


_USEQ_AXIS_TO_PYCRO: dict[str, PycroAxis] = {
    Axis.TIME: "time",
    Axis.POSITION: "position",
    Axis.Z: "z",
    Axis.CHANNEL: "channel",
}
_USEQ_KEY_TO_PYCRO: dict[str, PycroKey] = {
    "exposure": "exposure",
    "x_pos": "x",
    "y_pos": "y",
    "z_pos": "z",
    "min_start_time": "min_start_time",
    "keep_shutter_open": "keep_shutter_open",
}


def _event_to_pycromanager(event: MDAEvent) -> PycroManagerEvent:
    """Convenience method to convert this event to a pycro-manager events.

    See: <https://pycro-manager.readthedocs.io/en/latest/apis.html>
    """
    pycro: PycroManagerEvent = {"axes": {}}

    for axis in event.index.keys():
        if axis in _USEQ_AXIS_TO_PYCRO:
            pycro["axes"][_USEQ_AXIS_TO_PYCRO[axis]] = event.index[axis]

    for useq_name, pycro_name in _USEQ_KEY_TO_PYCRO.items():
        if (val := getattr(event, useq_name)) is not None:
            pycro[pycro_name] = val

    if event.channel:
        pycro["config_group"] = [event.channel.group, event.channel.config]

    if event.properties:
        pycro["properties"] = [list(p) for p in event.properties]

    return pycro



from __future__ import annotations

import math
from typing import Any, Callable, Iterator, List, Sequence, Union

import numpy as np

from useq._base_model import FrozenModel
from useq._pydantic_compat import field_validator


def _list_cast(field: str) -> Callable[[Any], Any]:
    v = field_validator(field, mode="before", allow_reuse=True, check_fields=False)
    return v(list)


class ZPlan(FrozenModel):
    go_up: bool = True

    def __iter__(self) -> Iterator[float]:  # type: ignore
        positions = self.positions()
        if not self.go_up:
            positions = positions[::-1]
        yield from positions

    def _start_stop_step(self) -> tuple[float, float, float]:
        raise NotImplementedError

    def positions(self) -> Sequence[float]:
        start, stop, step = self._start_stop_step()
        return list(np.arange(start, stop + step, step))

    def num_positions(self) -> int:
        start, stop, step = self._start_stop_step()
        return math.ceil((stop + step - start) / step)

    @property
    def is_relative(self) -> bool:
        return True


class ZTopBottom(ZPlan):
    """Define Z using absolute top & bottom positions.

    Note that `bottom` will always be visited, regardless of `go_up`, while `top` will
    always be *encompassed* by the range, but may not be precisely visited if the step
    size does not divide evenly into the range.

    Attributes
    ----------
    top : float
        Top position (inclusive).
    bottom : float
        Bottom position (inclusive).
    step : float
        Step size in microns.
    go_up : bool
        If `True`, instructs engine to start at bottom and move towards top. By default,
        `True`.
    """

    top: float
    bottom: float
    step: float

    def _start_stop_step(self) -> tuple[float, float, float]:
        return self.bottom, self.top, self.step

    @property
    def is_relative(self) -> bool:
        return False


class ZRangeAround(ZPlan):
    """Define Z as a symmetric range around some reference position.

    Note that `-range / 2` will always be visited, regardless of `go_up`, while
    `+range / 2` will always be *encompassed* by the range, but may not be precisely
    visited if the step size does not divide evenly into the range.

    Attributes
    ----------
    range : float
        Range in microns (inclusive). For example, a range of 4 with a step size
        of 1 would visit [-2, -1, 0, 1, 2].
    step : float
        Step size in microns.
    go_up : bool
        If `True`, instructs engine to start at bottom and move towards top. By default,
        `True`.
    """

    range: float
    step: float

    def _start_stop_step(self) -> tuple[float, float, float]:
        return -self.range / 2, self.range / 2, self.step


class ZAboveBelow(ZPlan):
    """Define Z as asymmetric range above and below some reference position.

    Note that `below` will always be visited, regardless of `go_up`, while `above` will
    always be *encompassed* by the range, but may not be precisely visited if the step
    size does not divide evenly into the range.

    Attributes
    ----------
    above : float
        Range above reference position in microns (inclusive).
    below : float
        Range below reference position in microns (inclusive).
    step : float
        Step size in microns.
    go_up : bool
        If `True`, instructs engine to start at bottom and move towards top. By default,
        `True`.
    """

    above: float
    below: float
    step: float

    def _start_stop_step(self) -> tuple[float, float, float]:
        return -abs(self.below), +abs(self.above), self.step


class ZRelativePositions(ZPlan):
    """Define Z as a list of positions relative to some reference.

    Typically, the "reference" will be whatever the current Z position is at the start
    of the sequence.

    Attributes
    ----------
    relative : list[float]
        List of relative z positions.
    go_up : bool
        If `True` (the default), visits points in the order provided, otherwise in
        reverse.
    """

    relative: List[float]

    _normrel = _list_cast("relative")

    def positions(self) -> Sequence[float]:
        return self.relative

    def num_positions(self) -> int:
        return len(self.relative)


class ZAbsolutePositions(ZPlan):
    """Define Z as a list of absolute positions.

    Attributes
    ----------
    relative : list[float]
        List of relative z positions.
    go_up : bool
        If `True` (the default), visits points in the order provided, otherwise in
        reverse.
    """

    absolute: List[float]

    _normabs = _list_cast("absolute")

    def positions(self) -> Sequence[float]:
        return self.absolute

    def num_positions(self) -> int:
        return len(self.absolute)

    @property
    def is_relative(self) -> bool:
        return False


# order matters... this is the order in which pydantic will try to coerce input.
# should go from most specific to least specific
AnyZPlan = Union[
    ZTopBottom, ZAboveBelow, ZRangeAround, ZAbsolutePositions, ZRelativePositions
]



# don't add __future__.annotations here
# pydantic2 isn't rebuilding the model correctly

import warnings
from types import MappingProxyType
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
)

from pydantic import Field

from useq._actions import AcquireImage, AnyAction
from useq._base_model import UseqModel
from useq._pydantic_compat import PYDANTIC2, field_serializer

if TYPE_CHECKING:
    from useq._mda_sequence import MDASequence
    from useq.pycromanager import PycroManagerEvent

    ReprArgs = Sequence[Tuple[Optional[str], Any]]


class Channel(UseqModel):
    """Channel in a MDA event.

    Attributes
    ----------
    config : str
        Name of the configuration to use for this channel, (e.g. `"488nm"`, `"DAPI"`,
        `"FITC"`).
    group : str
        Optional name of the group to which this channel belongs. By default,
        `"Channel"`.
    """

    config: str
    group: str = "Channel"

    def __eq__(self, _value: object) -> bool:
        if isinstance(_value, str):
            return self.config == _value
        return super().__eq__(_value)


class PropertyTuple(NamedTuple):
    """Three-tuple capturing a device, property, and value.

    Attributes
    ----------
    device_name : str
        Name of a device.
    property_name : str
        Name of a property recognized by the device.
    value : Any
        Value for the property.
    """

    device_name: str
    property_name: str
    property_value: Any


def _float_or_none(v: Any) -> Optional[float]:
    return float(v) if v is not None else v


class MDAEvent(UseqModel):
    """Define a single event in a [`MDASequence`][useq.MDASequence].

    Usually, this object will be generator by iterating over a
    [`MDASequence`][useq.MDASequence] (see [`useq.MDASequence.iter_events`][]).

    Attributes
    ----------
    index : dict[str, int]
        Index of this event in the sequence. This is a mapping of axis name
        to index.  For example: `{'t': 4, 'c': 0, 'z': 5},`
    channel : Channel | None
        Channel to use for this event. If `None`, implies use current channel.
        By default, `None`.  `Channel` is a simple pydantic object with two attributes:
        `config` and `group`.  `config` is the name of the configuration to use for this
        channel, (e.g. `"488nm"`, `"DAPI"`, `"FITC"`).  `group` is the name of the group
        to which this channel belongs. By default, `"Channel"`.
    exposure : float | None
        Exposure time in milliseconds. If not provided, implies use current exposure
        time. By default, `None`.
    min_start_time : float | None
        Minimum start time of this event, in seconds.  If provided, the engine will
        pause until this time has elapsed (relative to the start of the sequence)
        before starting this event. By default, `None`.
    pos_name : str | None
        The name assigned to the position. By default, `None`.
    x_pos : float | None
        X position in microns. If not provided, implies use current position. By
        default, `None`.
    y_pos : float | None
        Y position in microns. If not provided, implies use current position. By
        default, `None`.
    z_pos : float | None
        Z position in microns. If not provided, implies use current position. By
        default, `None`.
    sequence : MDASequence | None
        A reference to the [`useq.MDASequence`][] this event belongs to. This is a
        read-only attribute. By default, `None`.
    properties : Sequence[PropertyTuple] | None
        List of [`useq.PropertyTuple`][] to set before starting this event. Where each
        item in the list is a 3-member named tuple of `(device_name, property_name,
        property_value)`.  This is inspired by micro-manager's Device Adapter API, but
        could be used to set arbitrary properties in any backend that supports the
        concept of devices that have properties with values. By default, `None`.
    metadata : dict
        Optional metadata to be associated with this event.
    action : Action
        The action to perform for this event.  By default, [`useq.AcquireImage`][].
        Example of another action is [`useq.HardwareAutofocus`][] which could be used
        to perform a hardware autofocus.
    keep_shutter_open : bool
        If True, the illumination shutter should be left open after the event has
        been executed, otherwise it should be closed. By default, `False`."
        This is useful when the sequence of events being executed use the same
        illumination scheme (such as a z-stack in a single channel), and closing and
        opening the shutter between events would be slow.
    """

    # MappingProxyType is not subscriptable on Python 3.8
    index: Mapping[str, int] = Field(default_factory=lambda: MappingProxyType({}))
    channel: Optional[Channel] = None
    exposure: Optional[float] = Field(default=None, gt=0.0)
    min_start_time: Optional[float] = None  # time in sec
    pos_name: Optional[str] = None
    x_pos: Optional[float] = None
    y_pos: Optional[float] = None
    z_pos: Optional[float] = None
    sequence: Optional["MDASequence"] = Field(default=None, repr=False)
    properties: Optional[List[PropertyTuple]] = None
    metadata: Dict[str, Any] = Field(default_factory=dict)
    action: AnyAction = Field(default_factory=AcquireImage)
    keep_shutter_open: bool = False

    @property
    def global_index(self) -> int:
        warnings.warn(
            "global_index is no longer functional.  Use `enumerate()` "
            "on an iterable of events instead.",
            stacklevel=2,
        )
        return 0

    def to_pycromanager(self) -> "PycroManagerEvent":
        from useq.pycromanager import to_pycromanager

        warnings.warn(
            "useq.MDAEvent.to_pycromanager() is deprecated and will be removed in a "
            "future version. Useq useq.pycromanager.to_pycromanager(event) instead.",
            FutureWarning,
            stacklevel=2,
        )

        return to_pycromanager(self)

    if PYDANTIC2:
        _si = field_serializer("index", mode="plain")(lambda v: dict(v))
        _sx = field_serializer("x_pos", mode="plain")(_float_or_none)
        _sy = field_serializer("y_pos", mode="plain")(_float_or_none)
        _sz = field_serializer("z_pos", mode="plain")(_float_or_none)



from __future__ import annotations

from typing import Optional

from pydantic import Field

from useq._base_model import FrozenModel


class Channel(FrozenModel):
    """Define an acquisition channel.

    Attributes
    ----------
    config : str
        Name of the configuration to use for this channel, (e.g. `"488nm"`, `"DAPI"`,
        `"FITC"`).
    group : str
        Optional name of the group to which this channel belongs. By default,
        `"Channel"`.
    exposure : float | None
        Exposure time in milliseconds. Must be positive.  If not provided, implies use
        current exposure time. By default, `None`.
    do_stack : bool
        If `True`, instructs engine to include this channel in any Z stacks being
        acquired. By default, `True`.
    z_offset : float
        Relative Z offset from current position, in microns. By default, `0`.
    acquire_every : int
        Acquire every Nth frame (if acquiring a time series). By default, `1`.
    camera: str | None
        Name of the camera to use for this channel. If not provided, implies use
        current camera. By default, `None`.
    """

    config: str
    group: str = "Channel"
    exposure: Optional[float] = Field(None, gt=0.0)
    do_stack: bool = True
    z_offset: float = 0.0
    acquire_every: int = Field(default=1, gt=0)  # acquire every n frames
    camera: Optional[str] = None



from typing import Union

from typing_extensions import Literal

from useq._base_model import FrozenModel


class Action(FrozenModel):
    """Base class for a [`useq.MDAEvent`][] action.

    An `Action` specifies what task should be performed during a
    [`useq.MDAEvent`][]. An `Action` can be for example used to acquire an
    image ([`useq.AcquireImage`][]) or to perform a hardware autofocus
    ([`useq.HardwareAutofocus`][]).

    Attributes
    ----------
    type : str
        Type of the action that should be performed at the [`useq.MDAEvent`][].
    """

    type: str


class AcquireImage(Action):
    """[`useq.Action`][] to acquire an image.

    Attributes
    ----------
    type : Literal["acquire_image"]
        This action can be used to acquire an image.
    """

    type: Literal["acquire_image"] = "acquire_image"


class HardwareAutofocus(Action):
    """[`useq.Action`][] to perform a hardware autofocus.

    See also [`useq.AutoFocusPlan`][].

    Attributes
    ----------
    type : Literal["hardware_autofocus"]
        This action can be used to trigger hardware autofocus.
    autofocus_device_name : str
        The name of the hardware autofocus device.
    autofocus_motor_offset: float
        Before autofocus is performed, the autofocus motor should be moved to this
        offset.
    max_retries : int
        The number of retries if autofocus fails. By default, 3.
    """

    type: Literal["hardware_autofocus"] = "hardware_autofocus"
    autofocus_device_name: str
    autofocus_motor_offset: float
    max_retries: int = 3


AnyAction = Union[HardwareAutofocus, AcquireImage]
